import urx
import socket
import time
import array
import numpy as np
from numpy.polynomial.polynomial import polyfit
from itertools import compress

#### GLOBAL VARIABLES ####
sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
conn = addr = rob = None


#### COMMUNICATION BETWEEN PROCESSES ####
def open_and_connect_socket():
    """
    Open a host socket and connect to Halcon (TCP) socket, and connect to UR robot
    """

    global conn, addr, sock, rob
    rob = urx.Robot("192.168.0.90")

    HOST = '192.168.0.214'
    #HOST = '127.0.0.1'  # Standard loopback interface address (localhost)
    PORT = 3000        # Port to listen on (non-privileged ports are > 1023)

    sock.bind((HOST, PORT))
    sock.listen()

    conn, addr = sock.accept()
    print('Connected by', addr)


def close_socket():
    """
    Close socket and reset connection and address variable
    """

    sock.close()
    global conn, addr
    conn, addr = None


def send_component_data(component):
    global conn
    conn.send(bytes(component, 'utf-8'))


def receive_coordinates():
    """
    receives coordinate found by Halcon. Must always be called after a send_component_data(component) coomand
    """
    global conn
    # Receive data to calculate how many bytes will be sent
    data = conn.recv(1000)
    value = int.from_bytes(data, "big") # Convert bytes to int
    necessary_bytes = 8*value
    conn.send(bytes("received", 'utf-8')) # Send "receipt"

    if necessary_bytes == 0:
        return 0, 0 # TODO: Develop procedure for dealing with this error (no points found by Halcon)

    # Receive row coordinates
    row_bytes = conn.recv(necessary_bytes)
    row_doubles = array.array('d', row_bytes) # Convert bytes to double
    row_doubles.byteswap()  # Swap byte order
    row_coordinates = row_doubles.tolist() # Convert to simple list
    conn.send(bytes("received", 'utf-8')) # Send "receipt"


    # Receive column coordinates
    column_bytes = conn.recv(necessary_bytes)
    column_doubles = array.array('d', column_bytes) # Convert bytes to double
    column_doubles.byteswap()  # Swap byte order
    column_coordinates = column_doubles.tolist() # Convert to simple list
    conn.send(bytes("received", 'utf-8')) # Send "receipt"

    coordinates = list(zip(row_coordinates, column_coordinates))
    return coordinates


#### ROBOT POSITIONS ####
def initial_position():
    global rob
    a = 0.1
    v = 0.2
    # initial position (place for gripping circuits)
    pose = [0.55000000, 0.1300000, 0.0000000, -3.0763000, 0.54400000000, 0.000000000]
    rob.movel(pose, acc=a, vel=v, wait=False)

def standby_position():
    global rob
    a = 0.1
    v = 0.2
    # position in standby (while not operating)
    pose = [0.55000000, 0.1300000, 0.0000000, -3.0763000, 0.54400000000, 0.000000000]
    rob.movel(pose, acc=a, vel=v, wait=False)

def initial_placement_position():
    # Put gripper in placement position (hardcoded)
    global rob
    a = 0.1
    v = 0.2

    # intermediary placement positions
    pose = [0.6400000000, 0.15300000000, -0.1720000000000, -2.232560000000, -2.164207000000, -0.0000000000]
    rob.movel(pose, acc=a, vel=v, wait=False)
    time.sleep(5)
    pose = [0.6400000000, 0.15300000000, -0.2377500000000, -2.232560000000, -2.164207000000, -0.0000000000]
    rob.movel(pose, acc=a, vel=v, wait=False)
    time.sleep(3)
    pose = [0.6304112089014159, 0.15427731424875984, -0.24339087122316824, -2.203402639143257, -2.1856337412595934, 0.00852759018218581]
    rob.movel(pose, acc=a, vel=v, wait=False)
    time.sleep(3)


def plane_high():
    # Operate gripper in a plane over the composite
    global rob
    a = 0.1
    v = 0.2
    pose = rob.getl()
    pose[2] = -0.2377500000000
    rob.movel(pose, acc=a, vel=v, wait=False)
    time.sleep(3)


def plane_low():
    # Operate gripper in a plane on the composite
    global rob
    a = 0.1
    v = 0.2
    pose = rob.getl()
    pose[2] = -0.243400000000000
    rob.movel(pose, acc=a, vel=v, wait=False)
    time.sleep(3)


#### ROBOT COMMANDS ####
def move_gripper(x, y, angle):
    # Move the robot tool center point in 2d, first move coordinate position, then move joint angle in radians
    global rob, np
    a = 0.1
    v = 0.2

    if (angle != 0):
        angle_rad = np.deg2rad(angle)
        posej = rob.getj()
        posej[5] += angle_rad
        rob.movej(posej, acc=a, vel=v, wait=False)
        time.sleep(2)

    if (x != 0 or y != 0):
        # Correcting for robot axis perpendicular on camera axis
        robot_x = y
        robot_y = x

        posel = rob.getl()
        posel[0] += robot_x
        posel[1] += robot_y
        rob.movel(posel, acc=a, vel=v, wait=False)
        time.sleep(2)


def close_gripper():
    global rob
    rob.set_digital_out(3,False)
    time.sleep(1)
    rob.set_digital_out(2,True)

def open_gripper():
    global rob
    rob.set_digital_out(2,False)
    time.sleep(1)
    rob.set_digital_out(3,True)


#### COORDINATE PROCESSING ####
def coordinate_processing(coordinates, component):
    # Find "false" coordinates and estimate missing coordinates. Halcon coordinate system has origin in top left corner,
    # with ascending values in downward direction and to the right. Halcon coordinates sorted by ascending row (y).
    # Sometimes this sorting isnt correct, because of the areas are too large and weird.

    # Remove duplicates
    coordinates = list(set(coordinates))

    loop_on = True
    x, y = zip(*coordinates)
    x = np.array(list(x))
    y = np.array(list(y))
    indices = np.arange(len(x))

    line_1 = None
    line_2 = None

    for i in indices:

        j = i + 1
        while loop_on:

            if j > len(x) - 1:
                break

            tempx = np.array([x[i], x[j]])
            tempy = np.array([y[i], y[j]])

            b, m = polyfit(tempx, tempy, 1) # fitting a line between two points
            dist1 = y-(b + m * x)
            dist1 = abs(dist1)

            error = 0.001
            fil = [i < error for i in dist1]
            if (sum(fil) > 3):
                # Checking that the line goes through at least 4 points (based on the assumption that noise will not lay itself in lines)
                # (this is however not certain, and can happen in certain edge cases)
                # TODO: Create a fix for edge cases, when noise lays itself on a line of more than 3 points
                if line_1 is None:
                    line_1 = [b,m]
                    line_1_fil = fil
                    indices = np.setdiff1d(indices, np.array([i,j]))
                    break
                if line_1 is not None and (b < line_1[0] - error or b > line_1[0] + error)  and (line_1[1] - 10*error < m < line_1[1] +  10*error):
                    # If the line found does not have the same y-intercept and the slope is roughly parallell, then this line should be stored
                    line_2 = [b,m]
                    line_2_fil = fil
                    loop_on = False
            j += 1

    # Select all points connected to each line
    x_1 = np.array(list(compress(x,line_1_fil)))
    y_1 = np.array(list(compress(y,line_1_fil)))

    x_2 = np.array(list(compress(x,line_2_fil)))
    y_2 = np.array(list(compress(y,line_2_fil)))

    # Calculating a fit based on all the apropriate points
    b_1, m_1 = polyfit(x_1, y_1, 1)
    b_2, m_2 = polyfit(x_2, y_2, 1)

    dist_1 = abs(y-(b_1 + m_1 * x))
    dist_2 = abs(y-(b_2 + m_2 * x))

    error = 0.001
    line_1_fil = [i < error for i in dist_1]
    line_2_fil = [i < error for i in dist_2]

    x_1 = np.array(list(compress(x,line_1_fil)))
    y_1 = np.array(list(compress(y,line_1_fil)))

    x_2 = np.array(list(compress(x,line_2_fil)))
    y_2 = np.array(list(compress(y,line_2_fil)))

    if b_1 > b_2:
        return [[b_1,m_1,x_1,y_1],[b_2,m_2,x_2,y_2]]
    elif b_2 > b_1:
        return [[b_2,m_2,x_2,y_2],[b_1,m_1,x_1,y_1]]
    elif b_1 == b_2:
        return [[b_1,m_1,x_1,y_1],[b_2,m_2,x_2,y_2]] #TODO: this shouldn't be possible, so perhaps add an error log or something




def align_composite_circuit(composite_processed, circuit_processed):
    # This method uses processed coordinate information to determine appropriate robot commands

    # Angles
    comp_high_angle = np.rad2deg(np.arctan(composite_processed[0][1]))
    comp_low_angle = np.rad2deg(np.arctan(composite_processed[1][1]))
    circ_high_angle = np.rad2deg(np.arctan(circuit_processed[0][1]))
    circ_low_angle = np.rad2deg(np.arctan(circuit_processed[1][1]))

    high_angle_deviation = circ_high_angle - comp_high_angle
    low_angle_deviation = circ_low_angle - comp_low_angle

    # Setting the angle deviation to be corrected to the smallest deviation. This is so the movements remain stable.
    if abs(high_angle_deviation) < abs(low_angle_deviation):
        min_angle_deviation = high_angle_deviation
    elif abs(low_angle_deviation) < abs(high_angle_deviation):
        min_angle_deviation = low_angle_deviation
    elif low_angle_deviation == high_angle_deviation:
        min_angle_deviation = high_angle_deviation

    # Correction if deviation is more than 0.1 degrees
    if abs(min_angle_deviation) > 0.1:
        if abs(min_angle_deviation) > 5.7:
            min_angle_deviation = np.sign(min_angle_deviation)*0.5 # This is roughly the maximal angle deviation possible for the large ciruit. Instead it will do a minor correction and try again TODO: Implement component specific max deviations
        move_gripper(0,0,-min_angle_deviation)
        return True

    # Y-intersects
    comp_high_line = composite_processed[0][0]
    comp_low_line = composite_processed[1][0]
    circ_high_line = circuit_processed[0][0]
    circ_low_line = circuit_processed[1][0]

    high_line_deviation = circ_high_line - comp_high_line
    low_line_deviation = circ_low_line - comp_low_line

    # Setting the line deviation to be corrected to the largest deviation. All line "heights" are negative since origin is in the topleft corner
    if abs(high_line_deviation) < abs(low_line_deviation):
        min_line_deviation = high_line_deviation
    elif abs(low_line_deviation) < abs(high_line_deviation):
        min_line_deviation = low_line_deviation
    elif low_line_deviation == high_line_deviation:
        min_line_deviation = high_line_deviation

    # Correction if deviation is more than 0.0001 meters
    if abs(min_line_deviation) > 0.0001:
        # X-coordinates
        comp_high_coord = composite_processed[0][3]
        comp_low_coord = composite_processed[1][3]
        circ_high_coord = circuit_processed[0][3]
        circ_low_coord = circuit_processed[1][3]

        new_high_coord = [y - min_line_deviation for y in circ_high_coord]
        new_low_coord = [y - min_line_deviation for y in circ_low_coord]

        high_line_fil = [False]*len(new_high_coord)

        for index, y in enumerate(new_high_coord):
            if (index > len(new_high_coord) - 1) or (index > len(comp_high_coord) - 1):
                break
            high_line_fil[index] = y > comp_high_coord[index] - 0.001 and y < comp_high_coord[index] + 0.001

        low_line_fil = [False]*len(new_low_coord)

        for index, y in enumerate(new_low_coord):
            if (index > len(new_low_coord) - 1) or (index > len(comp_low_coord) - 1):
                break
            low_line_fil[index] = y > comp_low_coord[index] - 0.001 and y < comp_low_coord[index] + 0.001


        if sum(high_line_fil) > 3 and sum(low_line_fil) > 3:
            move_gripper(0,-min_line_deviation,0)
            return True


    x_comp_high = np.sort(composite_processed[0][2])
    x_comp_low = np.sort(composite_processed[1][2])

    x_circ_high =np.sort(circuit_processed[0][2])
    x_circ_low = np.sort(circuit_processed[1][2])

    # This next bit requires a bit of hardcoding depending on components,
    # so to add new components they need to be hardcoded in.
    # The idea here is that if the furthermost points of either composite or circuit remain undetected,
    # we should remove it for
    i = 0
    j = 0
    min_high_side_deviation = 1000000
    high_side_deviation = x_circ_high[i] - x_comp_high[j]
    while abs(high_side_deviation) > 0.006:
        prev_dev = high_side_deviation
        j += 1
        high_side_deviation = x_circ_high[i] - x_comp_high[j]
        if abs(high_side_deviation) < abs(min_high_side_deviation):
            min_high_side_deviation = high_side_deviation
        if (prev_dev > high_side_deviation):
            i += 1
            j = 0
        if j == len(x_comp_high) or i == len(x_circ_high):
            # If all distances are above 6 mm (the actual distance could potentially be more than 6 mm)
            break
    high_side_deviation = min_high_side_deviation

    i = 0
    j = 0
    min_low_side_deviation = 1000000
    low_side_deviation = x_circ_low[i] - x_comp_low[j]
    while abs(low_side_deviation) > 0.006:
        prev_dev = low_side_deviation
        j += 1
        low_side_deviation = x_circ_low[i] - x_comp_low[j]
        if abs(low_side_deviation) < abs(min_low_side_deviation):
            min_low_side_deviation = low_side_deviation
        if (prev_dev > low_side_deviation):
            i += 1
            j = 0
        if j == len(x_comp_low) or i == len(x_circ_low):
            # If all distances are above 6 mm (the actual distance could potentially be more than 6 mm)
            break
    low_side_deviation = min_low_side_deviation

    if abs(high_side_deviation) < abs(low_side_deviation):
        min_side_deviation = high_side_deviation
    elif abs(low_side_deviation) < abs(high_side_deviation):
        min_side_deviation = low_side_deviation
    elif low_side_deviation == high_side_deviation:
        min_side_deviation = high_side_deviation

    if abs(min_side_deviation) > 0.00015:
        # Transform x command to lie on the current "composite" coordinate system
        #x_tilde = min_side_deviation*np.cos(circ_high_angle)
        #y_tilde = min_side_deviation*np.sin(circ_high_angle)

        move_gripper(-min_side_deviation,0,0)
        return True

    return False

#### AUTOMATED PROCESS ####
def automatization_manager(component):

    composite = component.split("+")[0].lower()
    circuit = component.split("+")[1].lower()

    initial_position()
    input("Grip circuit manually, then press enter to continue")
    # TODO: Develop technique to automatically grip circuit

    # Localize substrate soldering points
    send_component_data(composite)
    coord_composite = receive_coordinates()
    composite_processed = coordinate_processing(coord_composite,composite)

    initial_placement_position()

    # Calibrate for gripper error
    send_component_data(circuit)
    coord_circuit = receive_coordinates()
    circuit_processed = coordinate_processing(coord_circuit,circuit)

    plane_high()
    correcting = align_composite_circuit(composite_processed,circuit_processed)

    iter = 0

    while correcting:

        if iter > 7:
            break
        #  Move to points, and correcting

        plane_low()
        send_component_data(circuit)
        coord_circuit = receive_coordinates()
        circuit_processed = coordinate_processing(coord_circuit,circuit)
        print(circuit_processed)

        plane_high()
        correcting = align_composite_circuit(composite_processed,circuit_processed)

        iter += 1

    plane_low()
    send_component_data(circuit)
    coord_circuit = receive_coordinates()
    circuit_processed = coordinate_processing(coord_circuit,circuit)
    # TODO: Add verification of positioning success
    #print(circuit_processed)
